/*============================================================================
                BREADTH FIRST VISIT OF A TREE (iterative version)
Grammar for the input tree:  char ::= '0' | '1' | ... | '9'
                             tree ::= char | - char tree | (tree char tree)
Internal tree:               Tree = Num  |  Num x Tree  |  Tree x Num x Tree
                                   (Leaf)     (Unary)          (Binary)
Grammar for the output tree: tree ::= n | (n.tree) | (tree.n.tree)
The program uses the module for queues in the file `queue_module.cpp'.
The program uses a queue of trees.
The class queue is: class queue {public : cell* first; cell* last;};
=========================================================================== */
#include <iostream.h> 
#include <stdlib.h>   // file needed for the EXIT_SUCCESS and the EXIT_FAILURE
//----------------------------------------------------------------------------
//                    trees
typedef int Values;
enum Tag {leaftree, unarytree, binarytree};
struct tree {
       struct Leaf   {Values value;}; 
       struct Unary  {Values value1; tree* son;};
       struct Binary {Values value2; tree* left; tree* right;};
       union {Leaf l; Unary u; Binary b;};
       Tag tag;};
typedef tree* treeptr;

void printtree(treeptr t)
{switch (t->tag){
case leaftree: cout << t->l.value; break;
case unarytree:
        cout << "(" << t->u.value1 << "."; 
        printtree(t->u.son);
        cout << ")"; break;
case binarytree:
        cout << "("; printtree(t->b.left);
        cout << "." << t->b.value2 << ".";
        printtree(t->b.right);          
        cout << ")"; break;
default: cout << "Error in printtree!";
         exit(EXIT_FAILURE);}};
//----------------------------------------------------------------------------
//                    from characters to integers
int v(char c){return c-48;};
//----------------------------------------------------------------------------
//                    parsing trees
void parsetree(treeptr &t) {
char c; cin >> c; 
if ('0' <= c && c <= '9') 
            {t->tag=leaftree; t->l.value=v(c);} else
if (c=='-') {t->tag=unarytree; cin >> c; t->u.value1=v(c);
             treeptr t1=new tree;
             parsetree(t1); t->u.son=t1;} else
if (c=='(') {t->tag=binarytree;
             treeptr t1=new tree;
             parsetree(t1); t->b.left=t1;
             cin >> c; t->b.value2=v(c);
             treeptr t2=new tree;
             parsetree(t2); t->b.right=t2;
             cin >> c;}
else {cerr << "Error in parsetree!"; exit(EXIT_FAILURE);};};
//----------------------------------------------------------------------------
//                    queue of trees
typedef treeptr QueueValues;
#include "queue_module.cpp"
//----------------------------------------------------------------------------
//                    printing a queue of trees
void printqueue(queue q)
{cout << " <=";
 if ((q.first) == NULL) {cout << " empty_queue";} else
 {cell* pointer=q.first;
  while (pointer != NULL)
     {cout << " "; printtree(pointer->atom);
      pointer=pointer->next;};};
 cout << " <=";}
//----------------------------------------------------------------------------
void main()
{queue queue1; emptyqueue(queue1);
treeptr pt=new tree;
cout << "Input a tree according to the given grammar above, please.\n";
parsetree(pt); cout << "\nTree in input:\n";
printtree(pt); cout << "\n\nBreadth first visit of the tree:\n";
inqueue(pt,queue1);
while (isemptyqueue(queue1) == false)
  {outqueue(pt,queue1);
   switch(pt->tag){
   case leaftree:   cout << pt->l.value << " "; break;
   case unarytree:  cout << pt->u.value1 << " ";
                    inqueue(pt->u.son,queue1); break;
   case binarytree: cout << pt->b.value2 << " ";
                    inqueue(pt->b.left,queue1);
                    inqueue(pt->b.right,queue1); break;
                   } // end switch
   }; // end while
cout << "\n";
}
/* input:   output:
------------------------------------------------------------------------------
            Input a tree according to the grammar given above, please.
            (-779-9(654))

            Tree in input:
            ((7.7).9.(9.(6.5.4)))

            Breadth first visit of the tree:
            9 7 9 7 5 6 4
--------------------------------------------------------------------------- */
